"""
hskl

Utilities Module
   
"""

# Author: Qian Cao

# License: BSD 3-Clause

import numpy as np
import matplotlib.pyplot as plt
import matplotlib as mpl
from spectral import principal_components
import h5py
from scipy.io import loadmat # for older .mat files

def overlay(X,Y,fig_params={},ax_params={}):
    
    # Y==0 corresponds to alpha=0
    alphas = (Y>0).astype(float)*0.6
    cmap = plt.get_cmap('Set3')
    norm = mpl.colors.Normalize(vmin=1,vmax=np.max(Y))
    RGBA = cmap(norm(Y))
    RGBA[:,:,3] = alphas
    
    fig = plt.figure(**fig_params)
    ax1 = plt.imshow(np.mean(X,axis=2), cmap='gray',interpolation='nearest',**ax_params)
    plt.axis('off')
    ax2 = plt.imshow(RGBA, interpolation='nearest', **ax_params)
    # ax2 = plt.imshow(Y, cmap='Set3',alpha=0.6,interpolation='nearest',
    #                  vmin=1, vmax=np.max(Y), **ax_params)
    plt.axis('off')
    
    return fig, ax1, ax2

def save_overlay(overlay_tuple, path):
    overlay_tuple[0].savefig(path, bbox_inches='tight',pad_inches = 0)

# Power normalization
def normalize_channels(x):
    # Assume (X,Y,Channels)
    rms = np.sqrt(np.mean(x**2,axis=2))
    return x/rms[:,:,None]

# Principal Component Analysis Perserving Fixed Variance
def pca_fit(x):
    return principal_components(x)
    
def pca_apply(x,pc,variance=0.8):
    pc_reduced = pc.reduce(fraction=variance)
    return pc_reduced.transform(x), pc_reduced.eigenvalues

# Read MATLAB file
def read_mat(filepath,keyind=0):
    """MATLAB file reader
    filepath: path to .mat file
    keyind: array in the .mat file to retrieve
    """
    try:
        f = h5py.File(filepath,'r')
        if hasattr(keyind, '__iter__'): # keyind can be a list or integer
            mat = []
            for kk in f.keys():
                mat.append(f['/'+kk].value)
        else:
            mat = f['/'+list(f.keys())[0]][()]
        return mat
    except OSError:
        print("Unable to open file, trying legacy matlab file loader.")
        
        try:
            fdict = loadmat(filepath)
            
            fdictkeys = list(fdict.keys()) 
            for key in fdictkeys: 
                if key.startswith("__"): 
                    fdict.pop(key)
            
            fdictkeys = list(fdict.keys()) 
            mat = fdict[fdictkeys[keyind]]
            
            mat = np.swapaxes(mat,0,2) # compatibility with h5py-loaded array
            
            return mat
        except:
            print(".mat file not readable using h5py or scipy.")

if __name__ == "__main__":
    pass


# -*- coding: utf-8 -*-

"""
hskl

Demo Module

* sample data

Useful links:
    
    http://lesun.weebly.com/hyperspectral-data-set.html
    https://www.kaggle.com/c/ipsa-ma511/overview

"""

# Author: Qian Cao

# License: BSD 3-Clause

import requests
from zipfile import ZipFile
import os

from tqdm import tqdm

from skimage import io

def dl_url(url, path, chunk_size=128):
    """ 
    Download dataset.
    
    Parameters
    ----------
    url : string
        Location of dataset in URL.
    path : string
        Location where the dataset is saved.
    chunk_size : int, optional
        Chunk size for download. The default is 128.

    Returns
    -------
    None.

    """
    
    r = requests.get(url, stream=True)
    total_size_in_bytes = int(r.headers.get('content-length'))
    
    with open(path, 'wb') as f:
        
        # Note: tqdm does not work well with print.
        progress_bar = tqdm(total=total_size_in_bytes, unit='iB', 
                            unit_scale=True, position=0, leave=True)
        
        for chunk in r.iter_content(chunk_size=chunk_size):
            # percent = int(count * chunk_size * 100 / total_size_in_bytes)
            progress_bar.update(len(chunk))
            f.write(chunk)            
            
        progress_bar.close()
            
def unzip(source_zip, target_path):
    """
    

    Parameters
    ----------
    source_zip : string
        Source zip file.
    target_path : string
        Location to be extracted.

    Returns
    -------
    None.

    """
    with ZipFile(source_zip, 'r') as zip: 
        zip.printdir()  
        zip.extractall() 
        print('Extraction complete.')

def dl_hyrank(path):
    """ 
    HyRANK Dataset
    
    https://www2.isprs.org/commissions/comm3/wg4/hyrank/
    
    Warning: This dataset is >400 MB.

    Returns
    -------
    h5py path containing trainng and test data arrays

    """
    
    url = "https://zenodo.org/record/1222202/files/HyRANK_satellite.zip?download=1"
    fn = "HyRANK_satellite.zip"
    
    path_fn = os.path.join(path,fn)
    dl_url(url, path_fn, chunk_size=128)

    unzip(path_fn, path)
    
def load_hyrank(path):
    """
    Load HyRANK dataset

    Parameters
    ----------
    path : string
        Directory containining the extracted hyrank dataset folder ("PATH/HyRANK_satellite")

    Returns
    -------
    List of training and validation image arrays

    """
    
    # Image names
    hyrank_dirname = "HyRANK_satellite"
    train_dirname = "TrainingSet"
    validation_dirname = "ValidationSet"
    
    train_fn_list = ["Dioni.tif","Loukia.tif"]
    train_label_fn_list = ["Dioni_GT.tif","Loukia_GT.tif"]
    validation_fn_list = ["Erato.tif","Kirki.tif","Nefeli.tif"]
    
    # Read TIF images as numpy arrays and append to list
    train = []
    train_label = []
    validation = []
    
    for ind, img_fn in enumerate(train_fn_list):
        
        train.append(io.imread(os.path.join(hyrank_dirname,
                                            train_dirname,
                                            img_fn)))
        
        train_label.append(io.imread(os.path.join(hyrank_dirname,
                                            train_dirname,
                                            train_label_fn_list[ind])))
    
    for ind, img_fn in enumerate(validation_fn_list):
        validation.append(io.imread(os.path.join(hyrank_dirname,
                                                 validation_dirname,
                                                 img_fn)))
    
    return train, train_label, validation


# -*- coding: utf-8 -*-

"""
hskl

Base Module

I/O Convention:
    Input array, output array
    Input list, output list (except flatten_with_la)
   
"""

# Author: Qian Cao

# License: BSD 3-Clause

import numpy as np
from sklearn.utils.validation import check_is_fitted
from multimethod import multimethod

# Integer label to ignore in label image
LABEL_IGNORE = 0

@multimethod
def shapeXYZ(X: np.ndarray):
    """ Returns XYZ dimensions as tuple for use in unflatten
    """
    return X.shape

@multimethod
def shapeXYZ(Xs: list):
    """ Returns XY dimensions as tuple for use in unflatten
    """
    Xshape_list = []
    for ind in range(len(Xs)):
        Xshape_list.append(shapeXYZ(Xs[ind]))
    return Xshape_list

@multimethod
def flatten3(X: np.ndarray):
    """ Flattens a single image <np.ndarray> (3D to 2D)
    Returns original array if already flattened
    """
    if X.ndim == 3: # assumes (DimX, DimY, Channels)
        s = X.shape
        Y = np.reshape(X, newshape=(s[0]*s[1],s[2]), order='C')
    elif X.ndim == 2: # image considered to be already flat, ignore
        Y = X # this should be (n_samples, n_features)
    else:
        raise ValueError('X.ndim must for 3 or 2 (already flat)')
    return Y

@multimethod
def flatten2(X: np.ndarray):
    """ Flattens a single LABEL image <np.ndarray> (2D to 1D)
    Returns original array if already flat
    """
    if X.ndim == 2: # assumes (DimX, DimY, Channels)
        s = X.shape
        Y = np.reshape(X, newshape=(s[0]*s[1]), order='C')
    elif X.ndim == 1: # image considered to be already flat, ignore
        Y = X # this should be (n_samples, n_features)
    else:
        raise ValueError('X.ndim must for 2 or 1 (already flat)')
    return Y

@multimethod
def flatten2(X: np.ndarray, Y: np.ndarray):
    """ Flattens a single LABEL image <np.ndarray> (2D to 1D) with MASK Y
    Returns original array if already flat
    """
    
    X, Y = flatten2(X), flatten2(Y)
    
    ind = np.squeeze(Y!=LABEL_IGNORE)
    X = X[np.squeeze(Y!=LABEL_IGNORE)]
    return X

@multimethod
def flatten3(X: np.ndarray, Y: np.ndarray):
    """ Flattens image with mask defined by Y (3D to 2D)
    Returns masked array if already flat
    """
    if X.ndim == 3 and Y.ndim == 3: # assumes (DimX, DimY, Channels)
        X, Y = flatten3(X), flatten2(Y)
        X = X[np.squeeze(Y!=LABEL_IGNORE),:]
    elif X.ndim == 2 and Y.ndim == 1 and X.shape[0] == Y.shape[0]:
        X = X[np.squeeze(Y!=LABEL_IGNORE),:] # images already flattened
    else:
        raise ValueError('(X.ndim, Y.ndim) must for (3,2) or (2,1)')
    return X

@multimethod
def flatten3(Xs: list):
    """ Flattens list of images
    Applies flatten() to each element in the list
    """
    n_images_ = len(Xs)
    n_features_list = []
    images_flattened = []
    
    for ind in range(n_images_):
        X = flatten3(Xs[ind])
        images_flattened.append(X)
        n_features_list.append(X.shape[1])
    
    if not len(set(n_features_list))==1:
        raise ValueError("All arrays in input list must be have the same number of features.")
        
    return np.concatenate(images_flattened,axis=0)

@multimethod
def flatten2(Xs: list):
    """ Flattens list of LABEL images
    Applies flatten() to each element in the list
    """
    n_images_ = len(Xs)
    n_features_list = []
    images_flattened = []
    
    for ind in range(n_images_):
        X = flatten2(Xs[ind])
        images_flattened.append(X)
        n_features_list.append(X.shape[1])
    
    if not len(set(n_features_list))==1:
        raise ValueError("All arrays in input list must be have the same number of features.")
        
    return np.concatenate(images_flattened,axis=0)

@multimethod
def flatten_with_label(X: np.ndarray, Y: np.ndarray):
    """ Flattens image X with label Y
    """
    X, Y = flatten3(X), flatten2(Y)
    X = X[np.squeeze(Y!=LABEL_IGNORE),:]
    Y = Y[Y!=LABEL_IGNORE]
    
    return X, Y
    
@multimethod
def flatten_with_label(Xs: list, Ys: list):
    """ Flattens list of images X with list of label images Y
    """
    n_images_ = len(Xs)
    
    n_features_list = []
    X_flattened = []
    Y_flattened = []
    
    for ind in range(n_images_):
        X, Y = flatten_with_label(Xs[ind], Ys[ind])
        n_features_list.append(X.shape[1])
        X_flattened.append(X)
        Y_flattened.append(Y)
        
    if not len(set(n_features_list))==1:
        raise ValueError("All arrays in input list must be have the same number of features.")
        
    X = np.concatenate(X_flattened,axis=0)
    Y = np.concatenate(Y_flattened,axis=0)
    
    return X, Y

def sample_fraction_random(X, Y, frac):
    """ Samples a random fraction of rows from sklearn matrix for training
    """
    Ntotal = X.shape[0]
    Nselect = int(Ntotal*frac)
    indices = np.random.choice(Ntotal,Nselect,replace=False)
    
    X = X[indices,:]
    Y = Y[indices]
    
    return X, Y

@multimethod
def unflatten(X: np.ndarray, s: tuple):
    """ Unflattens image with shape defined by tuple s (1D to 2D)
    s denotes dimensions of the *INPUT* image
    len(s) == 3 : reshape to 2D label image
    len(s) == 2 : input is flattened image, ignore.
    """
    if len(s) == 3: # input array was hyperspectral image, X is 2D
        Y = np.squeeze(np.reshape(X, newshape=(s[0], s[1]), order='C'))
    elif len(s) == 2: # input array was sklearn matrix, no need to reshape, X is 1D
        Y = X
    else:
        raise ValueError("Incompatible dimension")
    return Y

@multimethod
def unflatten(X: np.ndarray, shapes: list):
    """ Unflattens images with shape defined by list of tuples s
    X is an array (1D), unflattened to 2D
    s denotes dimensions of the *INPUT* image
    len(s) == 3 : reshape to 2D label image
    len(s) == 2 : input is flattened image, ignore.
    """
    n_images_ = len(shapes)
    n_samples_list = [] # Number of sampled pixel/spectra for each image
    
    for ind in range(n_images_):
        s = shapes[ind]
        if len(s) == 3:
            n_samples_list.append(s[0]*s[1]) # image pixel list
        elif len(s) == 2:
            n_samples_list.append(s[0])
            
    X_list = []
    
    sinds = np.cumsum(np.array(n_samples_list)) # sample indices
    sinds = np.concatenate(([0],sinds))
    
    for ind in range(n_images_):
        s = shapes[ind]
        sslice = slice(sinds[ind],sinds[ind+1]) # sample slice
        X_list.append(unflatten(X[sslice],s))
    
    return X_list

class HyperspectralMixin:
    
    @classmethod
    def _fit(cls, est, X, Y, sample_fraction=None):
        """ Fit hyperspectral image to estimator
        X and Y may be np.ndarrays or list of np.ndarrays
        handles both flattened and original sklearn formats
        """
        
        # For now, only supervised estimators are considered
        
        X, Y = flatten_with_label(X,Y)
        
        # fit on partial data
        if sample_fraction != None:
            X, Y = sample_fraction_random(X, Y, sample_fraction)
        
        cls.n_features_ = X.shape[1]
        
        est = est.fit(X,Y)
        est.is_fitted_ = True
        
        return est
    
    @classmethod
    def _predict(cls, est, X, Y=None):
        """ Predict using estimator
        X and Y may be np.ndarrays or list of np.ndarrays
        handles both flattened and original sklearn formats
        """
        
        # For now, prediction with mask is not supported
        # crop input X to achieve same result as output
        check_is_fitted(est, 'is_fitted_')
        
        Xshape = shapeXYZ(X) # tuples or list of tuples
        X = flatten3(X)
        X = est.predict(X)
        X = unflatten(X, Xshape) # only Xshape[0] and Xshape[1] are used.
        
        return X


# -*- coding: utf-8 -*-

"""
hskl

Classification Module

Data I/O conventions
X: (DimX, DimY, NumSpectralChannels) <np.array>
Y: (DimX, DimY) <np.array,dtype=uint8>
   0 - ignored
   1 - first label
   2 - second label
   ...
   
"""

# Author: Qian Cao

# License: BSD 3-Clause

import numpy as np

from sklearn.utils import all_estimators
from sklearn.base import BaseEstimator, ClassifierMixin
from .base import flatten_with_label, HyperspectralMixin
from sklearn.metrics import accuracy_score

_sklearn_methods = dict(all_estimators(type_filter='classifier'))

_methods_list = ['AdaBoostClassifier', 'BaggingClassifier', 'BernoulliNB',
            'DecisionTreeClassifier','ExtraTreesClassifier',
            'GaussianNB','GradientBoostingClassifier',
            'KNeighborsClassifier','LinearDiscriminantAnalysis',
            'LinearSVC','LogisticRegression','MLPClassifier',
            'QuadraticDiscriminantAnalysis','RandomForestClassifier',
            'RidgeClassifier','SGDClassifier','SVC']

_parallel_methods = []

_methods = {k:_sklearn_methods[k] for k in _methods_list
            if k in _sklearn_methods}

def list_sklearn_methods():
    """List Available Classifier Methods in Scikit-Learn
    
    Returns
    -------
    list of classifier names

    """
    return list(_sklearn_methods.keys())

def list_methods():
    """List Available Classifier Methods Tested for Hyperspectral-sklearn
    
    Returns
    -------
    list of classifier names

    """
    return list(_methods.keys())

class HyperspectralClassifier(BaseEstimator, HyperspectralMixin, ClassifierMixin):
    """Hyperspectral Classifier
    
    Wrapper of classifiers in sklearn
    
    Parameters
    ----------
    method_name : str, default='RandomForest'
        Invokes correponding classification algorithm in scikit-learn.
    method_param : dict, default={}
        A parameter used for demonstation of how to pass and store paramters.

    """
    
    _estimator_type = "classifier"
    
    def __init__(self, method_name='RandomForest', method_params={}):
        
        self.__dict__.update(method_params)
        
        if (method_name in _methods):
            self.est = _methods[method_name](**method_params)
        else:   
            raise KeyError(" ".join(["Method ", method_name, " not found. \
                         Use list_methods() to view available classifiers."]))
                         
    def fit(self, X, Y, sample_fraction=None):
        """ Fit model to inputs and labels.
        """
        self.est = self._fit(self.est, X, Y, sample_fraction)
        
        return self

    def predict(self, X, Y = None):
        """ Use model to predict input labels.
        """
        Y = self._predict(self.est, X, Y)
        
        return Y
    
    def score(self, X, Y):
        """ Evaluate Classification accuracy
        Adapted from Classifier Mixin
        """
        sample_weight = None # TODO: Implement sample weighting
        
        X, Y = flatten_with_label(X, Y)
        return super().score(X, Y, sample_weight)
    
    def _more_tags(self):
        return {'requires_y': True}

if __name__ == "__main__":
    pass


import setuptools

with open("README.md", "r", encoding="utf-8") as fh:
    long_description = fh.read()

setuptools.setup(
    name="hskl",
    version="0.0.2",
    author="Qian Cao",
    author_email="qcao.dev@gmail.com",
    description="Hyperspectral image analysis with scikit-learn.",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/qiancao/hskl",
    packages=setuptools.find_packages(),
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: BSD License",
        "Operating System :: OS Independent",
    ],
    python_requires='>=3.6',
    install_requires=['numpy',
                      'scikit-learn',
                      'multimethod',
                      'matplotlib',
                      'spectral',
                      'h5py',
                      'scipy',
                      'tqdm',
                      'scikit-image']
)
